import json
import math
import urllib, urllib2

from streetview_image import StreetViewImage

EARTH_RADIUS = 6371000

class Point(object):
    def __init__(self, latitude, longitude):
        """A single point in the Earth defined by its latitude and longitude in
        degrees.

        """
        self._latitude = latitude
        self._longitude = longitude

    @classmethod
    def for_address(cls, address):
        """Returns a Point defined by its address or None if no or multiple
        result could be found by Google Geocoding API.

        """
        url_params = {'address': address, 'sensor': 'false'}
        response = urllib2.urlopen('http://maps.googleapis.com/maps/api/'
                                   'geocode/json?%s'
                                   % urllib.urlencode(url_params))
        output = json.loads(response.read())
        if output.get('status', '') == 'OK':
            if len(output['results']) == 1:
                latlng = output['results'][0]['geometry']['location']
                return cls(latlng['lat'], latlng['lng'])
        return None

    @staticmethod
    def deg_to_rad(deg):
        """Converts degress to radians."""
        return (deg * math.pi) / 180

    @staticmethod
    def rad_to_deg(rad):
        """Converts radians to degress."""
        return (rad * 180) / math.pi

    def latitude(self):
        """Return the latitude of the point."""
        return self._latitude

    def latitude_in_rad(self):
        """Return the latitude of the point in radians."""
        return Point.deg_to_rad(self._latitude)

    def longitude(self):
        """Returns the longitude of the point."""
        return self._longitude

    def longitude_in_rad(self):
        """Return the longitude of the point in radians."""
        return Point.deg_to_rad(self._longitude)

    def bearing_to(self, point):
        """Returns the bearing [0, 360] to another Point."""
        dlng = Point.deg_to_rad(point._longitude - self._longitude)
        y = math.sin(dlng) * math.cos(point.latitude_in_rad())
        x = math.cos(self.latitude_in_rad()) * \
            math.sin(point.latitude_in_rad()) - \
            math.sin(self.latitude_in_rad()) * \
            math.cos(point.latitude_in_rad()) * math.cos(dlng)
        return ((Point.rad_to_deg(math.atan2(y, x)) + 360) % 360)

class RoutePoint(object):
    def __init__(self, point, bearing):
        """A single point in a route defined by its latitude, longitude and
        bearing value.
        
        """
        self._point = point
        self._bearing = bearing

    def google_street_view_image(self, heading=0, pitch=0, width=640,
                                 height=640):
        """Returns its Google StreetViewImage.

        heading -- compass heading of the camera relative to the front of the
        vehicle. Accepted values are from 0 to 360 (both values indicating
        North, with 90 indicating East, and 180 South).

        pitch -- up or down angle of the camera relative to the vehicle.

        width -- width of the image in pixels.

        height -- height of the image in pixels.

        """
        return StreetViewImage(self._point.latitude(), self._point.longitude(),
                               (self._bearing + heading) % 360, pitch, width,
                               height)

class Route(object):
    def __init__(self, points):
        """A route defined by its ordered points."""
        self._route_points = []
        for i in range(len(points) - 1):
            self._route_points.append(
                RoutePoint(points[i], points[i].bearing_to(points[i + 1]))
            )
        if len(points) > 1:
            self._route_points.append(
                RoutePoint(points[-1], points[-2].bearing_to(points[-1]))
            )
        if len(points) == 1:
            self._route_points.append(RoutePoint(points[0], 0))

    @classmethod
    def between(cls, start_point, end_point, max_distance_between_points=20):
        """Returns a Route between two Points or None if no route could be found
        by Google Directions API.

        start_point -- start Point of the route.

        end_point -- end Point of the route.

        max_distance_between_points -- maximum distance of consecutive
        RoutePoints in meters.

        """
        corner_points = []
        points = []
        url_params = {
            'origin': '%s,%s' % (start_point.latitude(),
                                 start_point.longitude()),
            'destination': '%s,%s' % (end_point.latitude(),
                                      end_point.longitude()),
            'sensor': 'false'
        }
        response = urllib2.urlopen(
            'http://maps.googleapis.com/maps/api/directions/'
            'json?%s' % urllib.urlencode(url_params)
        )
        output = json.loads(response.read())
        if output.get('status', '') == 'OK':
            steps = output['routes'][0]['legs'][0]['steps']
            for step in steps:
                corner_points.append(Point(step['start_location']['lat'],
                                           step['start_location']['lng']))
            if steps:
                corner_points.append(Point(steps[-1]['end_location']['lat'],
                                           steps[-1]['end_location']['lng']))
            for i in range(len(corner_points) - 1):
                points.append(corner_points[i])
                dlng = corner_points[i + 1].longitude_in_rad() - \
                       corner_points[i].longitude_in_rad()
                dist = math.acos(
                    math.sin(corner_points[i].latitude_in_rad()) *
                    math.sin(corner_points[i + 1].latitude_in_rad()) +
                    math.cos(corner_points[i].latitude_in_rad()) *
                    math.cos(corner_points[i + 1].latitude_in_rad()) * math.cos(dlng)
                ) * EARTH_RADIUS
                num_of_inter_points = int(
                    math.floor(dist / max_distance_between_points)
                )
                for j in range(1, num_of_inter_points + 1):
                    points.append(
                        Point(
                            Point.rad_to_deg(
                                corner_points[i].latitude_in_rad() +
                                ((1.0 * j) / (num_of_inter_points + 1)) *
                                (corner_points[i + 1].latitude_in_rad() -
                                 corner_points[i].latitude_in_rad())
                            ),
                            Point.rad_to_deg(
                                corner_points[i].longitude_in_rad() +
                                ((1.0 * j) / (num_of_inter_points + 1)) *
                                (corner_points[i + 1].longitude_in_rad() -
                                 corner_points[i].longitude_in_rad())
                            )
                        )
                    )
            if corner_points:
                points.append(corner_points[-1])
            return cls(points)
        return None

    def google_street_view_images(self, heading=0, pitch=0, width=640,
                                  height=640):
        """Returns its Google StreetViewImages.

        heading - compass heading of the camera relative to the front of the
        vehicle. Accepted values are from 0 to 360 (both values indicating
        North, with 90 indicating East, and 180 South).

        pitch - up or down angle of the camera relative to the vehicle.

        width - width of the image in pixels.

        height - height of the image in pixels.

        """
        return [
            route_point.google_street_view_image(heading, pitch, width, height)
            for route_point in self._route_points
        ]
