import Image
import math
import random
import urllib2

from cStringIO import StringIO

class Color(object):
    def __init__(self, r, g, b):
        """Count of RGB color in image.

        r -- red component.
        g -- green component.
        b -- blue component.
        
        """
        self._r = r
        self._g = g
        self._b = b

    @classmethod
    def medium_color(cls, colors):
        """Returns the medium color for a list of Colors.

        colors -- a list of Colors.
        
        """
        r = g = b = 0
        for color in colors:
            r += color._r
            g += color._g
            b += color._b
        n_colors = max(1, len(colors))
        return cls(r / n_colors, g / n_colors, b / n_colors)

    def red(self):
        """Returns its red component."""
        return self._r

    def green(self):
        """Returns its green component."""
        return self._g

    def blue(self):
        """Returns its blue component."""
        return self._b

    def euclidean_distance(self, color):
        """Returns the euclidean distance to other Color.
        
        color -- other Color to compute euclidean distance.
        
        """
        return math.sqrt(
            (self._r - color._r) ** 2 +
            (self._g - color._g) ** 2 +
            (self._b - color._b) ** 2
        )

    def hex_code(self):
        """Returns its hex code."""
        return '#%02x%02x%02x' % (self._r, self._g, self._b)

class StreetViewImage(object):
    def __init__(self, latitude, longitude, heading, pitch, width, height):
        """A Google StreetView image.

        latitude -- latitude of the point in degrees.

        longitude -- longitude of the point in degrees.

        heading -- compass heading of the camera relative to the front of the
        vehicle. Accepted values are from 0 to 360 (both values indicating
        North, with 90 indicating East, and 180 South).

        pitch -- up or down angle of the camera relative to the vehicle.

        width -- width of the image in pixels.

        height -- height of the image in pixels.

        """
        self._latitude = latitude
        self._longitude = longitude
        self._heading = heading
        self._pitch = pitch
        self._width = width
        self._height = height

    def url(self, key):
        """URL of the image.

        key -- identifies your application for quota purposes in Google APIs.

        """
        return 'http://maps.googleapis.com/maps/api/streetview' \
               '?size=%sx%s&location=%s,%s&sensor=false&heading=%s&pitch=%s' \
               '&key=%s' % (self._width, self._height, self._latitude,
                            self._longitude, self._heading, self._pitch, key)

    def dominant_colors(self, key, num_of_colors, x_offset=0, y_offset=0):
        """Returns a list of its dominant colors.

        key -- identifies your application for quota purposes in Google APIs.
        num_of_colors -- number of dominant colors to be calculated.
        x_offset -- crop left and right by x_offset pixels.
        y_offset -- crop top and bottom by y_offset pixels.

        """
        response = urllib2.urlopen(self.url(key))
        if response.getcode() != 200:
            raise Exception
        img = Image.open(StringIO(response.read()))
        width, height = img.size
        img = img.crop((0, y_offset, width, height - y_offset))
        colors = []
        for count, rgb in img.getcolors(width * height):
            colors.extend([Color(rgb[0], rgb[1], rgb[2])] * count)
        clusters = [[color, []]
                    for color in random.sample(colors, num_of_colors)]
        while True:
            for cluster in clusters:
                cluster[1] = []
            for color in colors:
                best_cluster = clusters[0]
                for cluster in clusters:
                    if color.euclidean_distance(cluster[0]) < \
                       color.euclidean_distance(best_cluster[0]):
                        best_cluster = cluster
                best_cluster[1].append(color)
            max_dif = 0
            for cluster in clusters:
                center = Color.medium_color(cluster[1])
                max_dif = max(max_dif, center.euclidean_distance(cluster[0]))
                cluster[0] = center
            if max_dif < 5:
                break
        return [cluster[0] for cluster in clusters]
